@@ -46,6 +46,7 @@ PER_MODULE_OPTIONS: Final = {
     "mypyc",
     "no_implicit_optional",
     "show_none_errors",
+    "strict_concatenate",
     "strict_equality",
     "strict_optional",
     "strict_optional_whitelist",
@@ -183,6 +184,9 @@ class Options:
         # This makes 1 == '1', 1 in ['1'], and 1 is '1' errors.
         self.strict_equality = False
 
+        # Make arguments prepended via Concatenate be truly positional-only.
+        self.strict_concatenate = False
+
         # Report an error for any branches inferred to be unreachable as a result of
         # type analysis.
         self.warn_unreachable = False
@@ -258,6 +262,7 @@ class Options:
         self.dump_type_stats = False
         self.dump_inference_stats = False
         self.dump_build_stats = False
+        self.enable_incomplete_features = False
 
         # -- test options --
         # Stop after the semantic analysis phase
